#!/usr/bin/env python

"""
Conversion functions and data used by XML-RPC and SOAP.

Copyright (C) 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free
Software Foundation; either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import datetime

# Utility classes.

class ParameterName(object):

    "A method parameter name."

    def __init__(self, ns, name):
        self.ns = ns
        self.name = name

    def __eq__(self, other):
        other_ns, other_name = other
        return self.ns, to_localName(self.name) == other_ns, to_localName(other_name)

    def __hash__(self):
        return hash(self.ns + to_localName(self.name))

    def __repr__(self):
        return "ParameterName(%s, %s)" % (repr(self.ns), repr(self.name or None))

    # Sequence emulation.

    def __len__(self):
        return 2

    def __getitem__(self, i):
        return (self.ns, self.name)[i]

def to_localName(name):
    return (name or "").split(":")[-1] or None

def convert(parameters, converters=None):

    """
    Convert the 'parameters', returning a list of name, value items where the
    names are the plain names for each parameter, and the values may be
    converted from strings to other data types.
    """

    conv = default_converters
    conv.update(converters or {})
    results = []
    for parameter_name, parameter_value in parameters:
        typename, name = parameter_name
        localName = to_localName(name)
        if isinstance(parameter_value, list):
            value = convert(parameter_value, converters)
        else:
            functions = conv.get(typename, {})
            function = functions.get(localName) or functions.get(None, unicode)
            value = function(parameter_value)
        results.append((localName, value))
    return results

# Utility functions.

def boolean(s):
    if s.lower() == "true":
        return True
    elif s.lower() == "false":
        return False
    else:
        raise ValueError, "String value %s not convertable to boolean." % repr(s)

def iso8601(s):
    # NOTE: To be written.
    return s

default_converters = {
    "string" : {None : unicode},
    "int" : {None : int},
    "i4" : {None : int},
    "double" : {None : float},
    "boolean" : {None : boolean},
    "dateTime.iso8601" : {None : iso8601},
    "base64" : {None : str}
    }

# vim: tabstop=4 expandtab shiftwidth=4
